@@ -1,15 +1,18 @@
 # coding: utf-8
-__version__ = "0.9.0"
+__version__ = "0.10.0"
+
 
 from .oep_client import OepClient
+from .advanced_api import AdvancedApiSession
 from .oep_client import (
     DEFAULT_HOST,
+    DEFAULT_PROTOCOL,
     DEFAULT_API_VERSION,
     DEFAULT_SCHEMA,
     DEFAULT_INSERT_RETRIES,
     DEFAULT_BATCH_SIZE,
 )
-from .oep_client import (
+from .exceptions import (
     OepApiException,
     OepServerSideException,
     OepClientSideException,
@@ -17,3 +20,4 @@ from .oep_client import (
     OepTableNotFoundException,
     OepTableAlreadyExistsException,
 )
+from .dialect import get_sqlalchemy_table
