@@ -1,10 +1,8 @@
 """Small client implementation to use the oep API.
 
-Version: 0.1
-
 Example usage: create a table, insert data, retrieve data, delete table
 
-cli = OEPClient(token='xxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxx')
+cli = OepClient(token='xxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxx')
 
 table = 'my_awesome_test_table'
 definition = {
@@ -35,40 +33,22 @@ import functools
 import click
 import requests
 
+from .exceptions import (
+    OepServerSideException,
+    OepClientSideException,
+    OepAuthenticationException,
+    OepTableNotFoundException,
+    OepTableAlreadyExistsException,
+)
+from .advanced_api import AdvancedApiSession
+from .dialect import get_sqlalchemy_table
 
-DEFAULT_HOST = "https://openenergy-platform.org"
+DEFAULT_HOST = "openenergy-platform.org"
+DEFAULT_PROTOCOL = "https"
 DEFAULT_API_VERSION = "v0"
 DEFAULT_SCHEMA = "model_draft"
-DEFAULT_BATCH_SIZE = None
-DEFAULT_INSERT_RETRIES = 0
-
-
-class OepApiException(Exception):
-    pass
-
-
-class OepServerSideException(OepApiException):
-    pass
-
-
-class OepClientSideException(OepApiException):
-    pass
-
-
-class OepAuthenticationException(OepClientSideException):
-    pass
-
-
-class OepTableNotFoundException(OepClientSideException):
-    def __init__(self, _msg):
-        # the API falsely returns message: {'detail': 'You do not have permission to perform this action}
-        # but this is only because table  does not exist
-        # TODO: create better error message on server side!
-        super().__init__("Table does not exist OR you don't have permission")
-
-
-class OepTableAlreadyExistsException(OepClientSideException):
-    pass
+DEFAULT_BATCH_SIZE = 5000
+DEFAULT_INSERT_RETRIES = 10
 
 
 def fix_table_definition(definition):
@@ -109,6 +89,7 @@ class OepClient:
     def __init__(
         self,
         token=None,
+        protocol=DEFAULT_PROTOCOL,
         host=DEFAULT_HOST,
         api_version=DEFAULT_API_VERSION,
         default_schema=DEFAULT_SCHEMA,
@@ -118,7 +99,8 @@ class OepClient:
         """
         Args:
             token(str): your API token
-            host(str, optional): host of the oep platform. default is "https://openenergy-platform.org"
+            host(str, optional): default is "https". "http" may be used for local installations
+            host(str, optional): host of the oep platform. default is "openenergy-platform.org"
             api_version(str, optional): currently only "v0"
             default_schema(str, optional): the default schema for the tables, usually "model_draft"
             batch_size(int, optional): number of records that will be uploaded per batch.
@@ -126,7 +108,10 @@ class OepClient:
             insert_retries(int, optional): number of insert_retries for insert on OepServerSideExceptions
         """
         self.headers = {"Authorization": "Token %s" % token} if token else {}
-        self.api_url = "%s/api/%s/" % (host, api_version)
+        self.api_url = "%s://%s/api/%s/" % (protocol, host, api_version)
+        self.protocol = protocol
+        self.host = host
+        self.token = token
         self.default_schema = default_schema
         self.batch_size = batch_size
         self.insert_retries = insert_retries
@@ -261,11 +246,8 @@ class OepClient:
                 * 'advanced' (default): sent records via advanced API
         """
 
-        table_def = self._get_table_definition(table, schema=schema)
-        column_names = [
-            c["name"]
-            for c in sorted(table_def["columns"], key=lambda c: c["ordinal_position"])
-        ]
+        table_def = self.get_table_definition(table, schema=schema)
+        column_names = [c["name"] for c in table_def["columns"]]
         used_column_names = set()
         for row in data:
             used_column_names = used_column_names | set(row.keys())
@@ -275,7 +257,7 @@ class OepClient:
                 "Columns not in table: %s", unknown_column_names
             )
 
-        batch_size = (batch_size or self.batch_size) or 1
+        batch_size = batch_size or self.batch_size
         n_batches = math.ceil(len(data) / batch_size)
         data_batches = [
             data[i * batch_size : (i + 1) * batch_size] for i in range(n_batches)
@@ -336,7 +318,7 @@ class OepClient:
         return res
 
     @check_exception("not found", OepTableNotFoundException)
-    def _get_table_definition(self, table, schema=None):
+    def get_table_definition(self, table, schema=None):
         """Returns table info
 
         Args:
@@ -348,13 +330,77 @@ class OepClient:
         """
         url = self._get_table_url(table=table, schema=schema)
         res = self._request("GET", url, 200)
+        definition = {
+            "columns": [],
+            "constraints": [],
+        }
+
+        # res["columns"] is dict
+        # first: add constrains to field definitions / constraints
+        for const in res["constraints"].values():
+            const_type = const["constraint_type"]
+            const_def = const["definition"]
+            if const_type == "PRIMARY KEY":
+                args = re.match(
+                    r"^PRIMARY KEY \((?P<field>[^)]+)\)$", const_def
+                ).groupdict()
+                # NOTE currently only single field PK allowed
+                res["columns"][args["field"]]["primary_key"] = True
+            elif const_type == "FOREIGN KEY":
+                args = re.match(
+                    r"^FOREIGN KEY \((?P<field>[^)]+)\) REFERENCES (?P<ref_schema>[^.]+)\.(?P<ref_table>[^()]+)\((?P<ref_field>[^)]+)\)$",
+                    const_def,
+                ).groupdict()
+                # NOTE currently only single field PK allowed
+                res["columns"][args["field"]]["foreign_key"] = [
+                    {
+                        "schema": args["ref_schema"],
+                        "table": args["ref_table"],
+                        "column": args["ref_field"],
+                    }
+                ]
+            elif const_type == "UNIQUE":
+                args = re.match(
+                    r"^UNIQUE \((?P<fields>[^)]+)\)$", const_def
+                ).groupdict()
+                columns = [f.strip() for f in args["fields"].split(",")]
+                definition["constraints"].append(
+                    {"constraint_type": "UNIQUE", "columns": columns}
+                )
+
+        def get_datatype(coldef):
+            dt = coldef["data_type"].upper()
+            if dt == "CHARACTER":
+                dt = "CHAR(%d)" % coldef["character_maximum_length"]
+            elif dt == "CHARACTER VARYING":
+                dt = "VARCHAR(%d)" % coldef["character_maximum_length"]
+            elif dt == "DOUBLE PRECISION":
+                dt = "FLOAT"
+
+            if (coldef["column_default"] or "").startswith("nextval"):
+                if not "INT" in dt:
+                    raise NotImplementedError(dt)
+                else:
+                    dt = dt.replace("INT", "SERIAL")
+
+            return dt
+
         # fix columns: list instead of dict
-        for name, col in res["columns"].items():
-            col["name"] = name
-        res["columns"] = sorted(
-            res["columns"].values(), key=lambda c: c["ordinal_position"]
-        )
-        return res
+        for name, coldef in sorted(
+            res["columns"].items(), key=lambda c: c[1]["ordinal_position"]
+        ):
+            col = {
+                "name": name,
+                "data_type": get_datatype(coldef),
+                "is_nullable": coldef["is_nullable"],
+            }
+            if coldef.get("primary_key", False):
+                col["primary_key"] = True
+            if "foreign_key" in coldef:
+                col["foreign_key"] = coldef["foreign_key"]
+            definition["columns"].append(col)
+
+        return definition
 
     def table_exists(self, table, schema=None):
         """True or False
@@ -366,12 +412,27 @@ class OepClient:
             schema(str, optional): table schema name.
                 defaults to self.default_schema which is usually "model_draft"
         """
+
         try:
-            self._get_table_definition(table, schema)
-            return True
+            return self._table_exists(table, schema)
         except OepTableNotFoundException:
             return False
 
+    @check_exception("not found", OepTableNotFoundException)
+    def _table_exists(self, table, schema=None):
+        """True or False
+
+        Args:
+            table(str): table name. Must be valid postgres table name,
+                all lowercase, only letters, numbers and underscore
+
+            schema(str, optional): table schema name.
+                defaults to self.default_schema which is usually "model_draft"
+        """
+        url = self._get_table_url(table=table, schema=schema)
+        self._request("GET", url, 200)
+        return True
+
     def get_metadata(self, table, schema=None):
         """Returns metadata json
 
@@ -440,11 +501,10 @@ class OepClient:
         }
         with self.advanced_session() as sas:
             # not data yet
-            res = sas._command("search", {"query": query})
+            res = sas._command("search", query)
             content = res["content"]
             description = content["description"]
             fieldnames = [f[0] for f in description]
-            _rowcount = content["rowcount"]  # should be 1
             res = sas._command("cursor/fetch_one")
             content = res["content"]
             rec = dict(zip(fieldnames, content))
@@ -458,79 +518,18 @@ class OepClient:
         )
         return self._request("POST", url, 200)
 
+    def get_sqlalchemy_table(self, table, schema=None):
+        return get_sqlalchemy_table(self, table, schema=schema)
 
-class AdvancedApiSession:
-    """Context for advanced api session (close connection on exit)"""
-
-    def __init__(self, oepclient):
-        """
-        Args:
-
-            oepclient(OEPClient)
-        """
-        self.oepclient = oepclient
-        self.api_url = self.oepclient.api_url + "advanced/"
-        self.connection_id = None
-        self.cursor_id = None
-
-    def _command(self, command, jsondata=None):
-        url = self.api_url + command
-        jsondata = jsondata or {}
-        if self.connection_id:
-            jsondata["connection_id"] = self.connection_id
-        if self.cursor_id:
-            jsondata["cursor_id"] = self.cursor_id
-        return self.oepclient._request("POST", url, 200, jsondata)
-
-    def __enter__(self):
-        self.connection_id = self._command("connection/open")["content"][
-            "connection_id"
-        ]
-        self.cursor_id = self._command("cursor/open")["content"]["cursor_id"]
-        logging.debug("Started connection: %s", self.connection_id)
-        return self
-
-    def __exit__(self, *args):
-        if self.cursor_id:
-            self._command("cursor/close")
-            logging.debug("Closed cursor: %s", self.cursor_id)
-            self.cursor_id = None
-        if self.connection_id:
-            self._command("connection/close")
-            logging.debug("Closed connection: %s", self.connection_id)
-            self.connection_id = None
-
-    def insert_into_table(self, table, data, schema):
-        """Insert records into table.
+    def delete_from_table(self, table, schema=None):
+        """Delete all rows from table (without dropping it).
 
         Args:
             table(str): table name. Must be valid postgres table name,
                 all lowercase, only letters, numbers and underscore
-            data(list): list of records(dict: column_name -> value)
-            schema(str): table schema name.
+            schema(str, optional): table schema name.
                 defaults to self.default_schema which is usually "model_draft"
+
         """
-
-        if not isinstance(data, (list, tuple)) or (
-            data and not isinstance(data[0], dict)
-        ):
-            raise OepClientSideException(
-                "data must be list or tuple of record dictionaries"
-            )
-
-        def _get_query(values):
-            query = {
-                "schema": schema or self.oepclient.default_schema,
-                "table": table,
-                "values": values,
-            }
-            return {"query": query}
-
-        try:
-            res = self._command("insert", _get_query(data))
-            self._command("connection/commit")
-            return res
-        except Exception as exc:
-            logging.error(exc)
-            self._command("connection/rollback")
-            raise
+        with self.advanced_session() as sas:
+            sas.delete_from_table(table, schema=schema)
